/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wNewton.h"
#include "wProblem.h"
#include "wFluid.h"
#include "wAssign.h"
#include "wFreestream.h"
#include "wWake.h"
#include "wKutta.h"
#include "wBlowing.h"
#include "wFpeLSFunction.h"
#include "wPotentialResidual.h"
#include "wFreestreamResidual.h"
#include "wWakeResidual.h"
#include "wKuttaResidual.h"
#include "wBlowingResidual.h"

#include "wMshData.h"
#include "wTag.h"
#include "wNode.h"
#include "wElement.h"
#include "wLinesearch.h"
#include "wLinearSolver.h"

#include <Eigen/Sparse>

#include <tbb/global_control.h>
#include <tbb/parallel_for_each.h>
#include <tbb/spin_mutex.h>

#include <iomanip>
#include <deque>

using namespace tbox;
using namespace dart;

#define ANSI_COLOR_RED "\x1b[1;31m"
#define ANSI_COLOR_GREEN "\x1b[1;32m"
#define ANSI_COLOR_YELLOW "\x1b[1;33m"
#define ANSI_COLOR_RESET "\x1b[0m"

/**
 * @brief Initialize the solver and set default parameters
 */
Newton::Newton(std::shared_ptr<Problem> _pbl, std::shared_ptr<tbox::LinearSolver> _linsol,
               double rTol, double aTol, int mIt,
               int nthrds, int vrb) : Solver(_pbl, _linsol, rTol, aTol, mIt, nthrds, vrb)
{
    // Additional display
    std::cout << "--- Newton Solver ---\n"
              << "Inner solver: " << *linsol
              << "Number of threads: " << nthreads << "\n"
              << "Relative tolerance: " << log10(relTol) << "\n"
              << "Absolute tolerance: " << log10(absTol) << "\n"
              << "Maximum iterations: " << maxIt << "\n"
              << std::endl;
    // Additional default parameters
    lsTol = 1e-6;
    maxLsIt = 10;
    avThrsh = 1e-2;
    mCOv = 1.;
    muCv = 1.;
}

/**
 * @brief Run the Newton solver
 *
 * Solve the steady transonic Full Potential Equation
 */
bool Newton::run()
{
    // Multithread
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);

    // Display current freestream conditions
    std::cout << std::setprecision(2)
              << "- Mach " << pbl->M_inf << ", "
              << pbl->alpha * 180 / 3.14159 << " deg AoA, "
              << pbl->beta * 180 / 3.14159 << " deg AoS"
              << std::endl;

    // Initialize solver loop
    nIt = 0;           // iteration counter
    int avC = 0;       // adaptive viscosity counter
    bool rUpwd = true; // update best upwind element

    // Residual and change in solution
    std::vector<double> dPhi(phi.size()); // dummy dPhi
    Eigen::Map<Eigen::VectorXd> rPhi_(rPhi.data(), rPhi.size()), dPhi_(dPhi.data(), dPhi.size());

    // Find initial best upwind element
    findUpwd();
    // Compute initial residual
    buildRes(rPhi_);
    double resInit = rPhi_.norm();
    double absRes = rPhi_.norm();
    double relRes = absRes / resInit;

    // Linesearch
    FpeLSFunction fpe(*this, phi, phi, dPhi_, rPhi_);
    BankRoseLS ls(fpe);
    //FleuryLS ls(fpe, 1., 1e-2, false);
    ls.set(maxLsIt, lsTol, verbose);

    // Display residual
    std::cout << std::setw(8) << "N_Iter"
              << std::setw(8) << "L_Iter"
              << std::setw(8) << "f_eval"
              << std::setw(12) << "Cl"
              << std::setw(12) << "Cd"
              << std::setw(15) << "Rel_Res[phi]"
              << std::setw(15) << "Abs_Res[phi]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(8) << nIt
              << std::setw(8) << 0
              << std::setw(8) << 1
              << std::setw(12) << "-"
              << std::setw(12) << "-"
              << std::setw(15) << log10(relRes)
              << std::setw(15) << log10(absRes) << "\n";

    do
    {
        // Viscosity damping
        if (relRes < avThrsh && avC < 2)
        {
            avC++;
            rUpwd = true;
        }
        switch (avC)
        {
        case 0:
        {
            mCOv = 0.92;
            muCv = 2.;
            break;
        }
        case 1:
        {
            mCOv = 0.935;
            muCv = 1.5;
            break;
        }
        default:
        {
            mCOv = 0.95;
            muCv = 1.;
            break;
        }
        }

        // Build Jacobian matrix
        Eigen::SparseMatrix<double, Eigen::RowMajor> J(pbl->msh->nodes.size(), pbl->msh->nodes.size());
        buildJac(J);

        // Solve linear SoE J \DeltaPhi = R
        tms["3-Slinsol"].start();
        linsol->compute(J, rPhi_, dPhi_);
        tms["3-Slinsol"].stop();
        nIt++;

        // Compute new solution and residual with linesearch
        tms["4-Rlnsrch"].start();
        fpe.update(phi, dPhi_);
        try
        {
            ls.run();
        }
        catch (...)
        {
        }
        tms["4-Rlnsrch"].stop();
        absRes = rPhi_.norm();
        relRes = absRes / resInit;
        // Re-find best upwind element
        if (rUpwd)
        {
            findUpwd();
            rUpwd = false;
        }

        // Compute aerodynamic coefficient on boundaries
        Solver::computeLoad();

        // Display residual (at each iteration)
        if (verbose > 0)
        {
            std::cout << std::fixed << std::setprecision(5);
            std::cout << std::setw(8) << nIt
                      << std::setw(8) << linsol->getIterations()
                      << std::setw(8) << ls.fevalIt
                      << std::setw(12) << Cl
                      << std::setw(12) << Cd
                      << std::setw(15) << log10(relRes)
                      << std::setw(15) << log10(absRes) << "\n";
        }

        // Check convergence
        if ((relRes < relTol || absRes < absTol) && nIt != 0)
            break;
        else if (std::isnan(relRes))
            break;
        else
            continue;
    } while (nIt < maxIt);

    // Display residual (only last iteration)
    if (verbose == 0)
    {
        std::cout << std::fixed << std::setprecision(5);
        std::cout << std::setw(8) << nIt
                  << std::setw(8) << linsol->getIterations()
                  << std::setw(8) << ls.fevalIt
                  << std::setw(12) << Cl
                  << std::setw(12) << Cd
                  << std::setw(15) << log10(relRes)
                  << std::setw(15) << log10(absRes) << "\n";
    }

    // Compute field variables
    Solver::computeFlow();

    // Check the solution
    if (relRes < relTol || absRes < absTol)
    {
        std::cout << ANSI_COLOR_GREEN << "Newton solver converged!" << ANSI_COLOR_RESET << std::endl;
        if (verbose > 1)
            std::cout << "Newton solver CPU" << std::endl
                      << tms;
        std::cout << std::endl;
        return true;
    }
    else if (std::isnan(relRes))
    {
        std::cout << ANSI_COLOR_RED << "Newton solver diverged!" << ANSI_COLOR_RESET << std::endl;
        if (verbose > 1)
            std::cout << "Newton solver CPU" << std::endl
                      << tms;
        std::cout << std::endl;
        return false;
    }
    else
    {
        std::cout << ANSI_COLOR_YELLOW << "Newton solver not fully converged!" << ANSI_COLOR_RESET << std::endl;
        if (verbose > 1)
            std::cout << "Newton solver CPU" << std::endl
                      << tms;
        std::cout << std::endl;
        return true;
    }
}

/**
 * @brief Build the Jacobian (tangent) matrix
 *
 * J = dR/dPhi = \int d( (1-mu)*rho + mu*rhoU * grad_phi ) / dPhi * grad_psi dV
 */
void Newton::buildJac(Eigen::SparseMatrix<double, Eigen::RowMajor> &J)
{
    // Multithread
    tbb::spin_mutex mutex;

    // List of triplets to build Jacobian matrix
    std::deque<Eigen::Triplet<double>> T;

    // Full Potential Equation with upwind bias: analytical derivatives
    tms["0-Jbase"].start();
    auto fluid = pbl->fluid;
    tbb::parallel_for_each(fluid->adjMap.begin(), fluid->adjMap.end(), [&](std::pair<Element *, std::vector<Element *>> p) {
        // Current element
        Element *e = p.first;
        //Upwind element
        Element *eU = p.second[0];
        // Build elementary matrices
        Eigen::MatrixXd Ae1, Ae2;
        std::tie(Ae1, Ae2) = PotentialResidual::buildGradientFlow(*e, *eU, phi, *fluid, muCv, mCOv);

        // Assembly (subsonic)
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t ii = 0; ii < e->nodes.size(); ++ii)
        {
            Node *nodi = e->nodes[ii];
            for (size_t jj = 0; jj < e->nodes.size(); ++jj)
            {
                Node *nodj = e->nodes[jj];
                T.push_back(Eigen::Triplet<double>(rows[nodi->row], nodj->row, Ae1(ii, jj)));
            }
        }
        // Assembly (supersonic)
        double mach = fluid->mach->eval(*e, phi, 0);
        if (mach > mCOv)
        {
            for (size_t ii = 0; ii < e->nodes.size(); ++ii)
            {
                Node *nodi = e->nodes[ii];
                for (size_t jj = 0; jj < eU->nodes.size(); ++jj)
                {
                    Node *nodj = eU->nodes[jj];
                    T.push_back(Eigen::Triplet<double>(rows[nodi->row], nodj->row, Ae2(ii, jj)));
                }
            }
        }
    });
    tms["0-Jbase"].stop();

    // Kutta condition step 2: equality of velocity norm through the wake
    // -> add the Kutta equation on the upper wake node rows
    // Nishida Thesis, pp. 29-30, Eq. 2.18
    tms["1-Jkutta"].start();
    for (auto wake : pbl->wBCs)
    {
        tbb::parallel_for_each(wake->wEle.begin(), wake->wEle.end(), [&](WakeElement *we) {
            // Build Ku, Kl
            Eigen::MatrixXd Aue, Ale;
            std::tie(Aue, Ale) = WakeResidual::buildGradientFlow(*we, phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < we->nRow; ++ii)
            {
                Node *nodi = we->wkN[ii].second;
                for (size_t jj = 0; jj < we->nColUp; ++jj)
                {
                    Node *nodj = we->volUpE->nodes[jj];
                    T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, Aue(ii, jj)));
                }
                for (size_t jj = 0; jj < we->nColLw; ++jj)
                {
                    Node *nodj = we->volLwE->nodes[jj];
                    T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, -Ale(ii, jj)));
                }
            }
        });
    }
    // Kutta condition step 3: equality of velocity norm on the Trailing edge (Kutta)
    // -> add the Kutta equation on the upper TE node rows (only works for 2D cases)
    // Galbraith paper, pp. 7, Eq. 34
    for (auto kutta : pbl->kSCs)
    {
        tbb::parallel_for_each(kutta->kEle.begin(), kutta->kEle.end(), [&](KuttaElement *ke) {
            // Build K
            Eigen::MatrixXd Ae = KuttaResidual::buildGradientFlow(*ke, phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < ke->nRow; ++ii)
            {
                Node *nodi = ke->teN[ii].second;
                for (size_t jj = 0; jj < ke->nCol; ++jj)
                {
                    Node *nodj = ke->volE->nodes[jj];
                    T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, Ae(ii, jj)));
                }
            }
        });
    }
    tms["1-Jkutta"].stop();

    // Build Jacobian matrix without BCs
    J.setFromTriplets(T.begin(), T.end());

    // Apply Dirichlet BCs
    tms["2-Jdbc"].start();
    for (auto dBC : pbl->dBCs)
    {
        for (auto nod : dBC->nodes)
        {
            for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(J, nod->row); it; ++it)
            {
                if (it.row() == it.col())
                    it.valueRef() = 1.;
                else
                    it.valueRef() = 0.;
            }
        }
    }
    tms["2-Jdbc"].stop();

    // Clean matrix and turn to compressed row format
    J.prune(0.);
    J.makeCompressed();

    if (verbose > 2)
        std::cout << "J (" << J.rows() << "," << J.cols() << ") nnz=" << J.nonZeros() << "\n";
}

/**
 * @brief Build the residual vector
 *
 * R = \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi dV
 */
void Newton::buildRes(Eigen::Map<Eigen::VectorXd> &R)
{
    // Multithread
    tbb::spin_mutex mutex;

    // Reset residual
    R.setZero();

    // Full Potential Equation with upwind bias
    auto fluid = pbl->fluid;
    tbb::parallel_for_each(fluid->adjMap.begin(), fluid->adjMap.end(), [&](std::pair<Element *, std::vector<Element *>> p) {
        // Current element
        Element *e = p.first;
        // Upwind element
        Element *eU = p.second[0];
        // Build elementary vector
        Eigen::VectorXd be = PotentialResidual::build(*e, *eU, phi, *fluid, muCv, mCOv);
        // Assembly (subsonic)
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t ii = 0; ii < e->nodes.size(); ++ii)
        {
            Node *nodi = e->nodes[ii];
            R(rows[nodi->row]) -= be(ii);
        }
    });
    // Apply freestream (Neumann) BCs
    for (auto nBC : pbl->fBCs)
    {
        tbb::parallel_for_each(nBC->tag->elems.begin(), nBC->tag->elems.end(), [&](Element *e) {
            // Build elementary flux vector
            Eigen::VectorXd be = FreestreamResidual::build(*e, phi, *nBC);

            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < e->nodes.size(); ++ii)
            {
                Node *nodi = e->nodes[ii];
                R(rows[nodi->row]) -= be(ii);
            }
        });
    }
    // Apply blowing (transpiration) BCs
    for (auto bBC : pbl->bBCs)
    {
        tbb::parallel_for_each(bBC->uE.begin(), bBC->uE.end(), [&](std::pair<Element *, F0ElC *> p) {
            // Build elementary flux vector
            Eigen::VectorXd be = BlowingResidual::build(*p.first, phi, *p.second);

            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < p.first->nodes.size(); ++ii)
            {
                Node *nodi = p.first->nodes[ii];
                R(rows[nodi->row]) -= be(ii);
            }
        });
    }
    // Slip BCs are naturally enforced in FEM, hence not applied explicitely

    // Kutta condition step 2: equality of velocity norm through the wake
    // -> add the Kutta equation on the upper wake node rows
    // Nishida Thesis, pp. 29-30, Eq. 2.18
    for (auto wake : pbl->wBCs)
    {
        tbb::parallel_for_each(wake->wEle.begin(), wake->wEle.end(), [&](WakeElement *we) {
            // Build bu, bl
            Eigen::VectorXd bue, ble;
            std::tie(bue, ble) = WakeResidual::build(*we, phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < we->nRow; ++ii)
            {
                Node *nodi = we->wkN[ii].second;
                R(nodi->row) -= bue(ii);
                R(nodi->row) += ble(ii);
            }
        });
    }
    // Kutta condition step 3: equality of velocity norm on the Trailing edge (Kutta)
    // -> add the Kutta equation on the upper TE node rows (only works for 2D cases)
    // Galbraith paper, pp. 7, Eq. 34
    for (auto kutta : pbl->kSCs)
    {
        tbb::parallel_for_each(kutta->kEle.begin(), kutta->kEle.end(), [&](KuttaElement *ke) {
            // Build b
            Eigen::VectorXd be = KuttaResidual::build(*ke, phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < ke->nRow; ++ii)
            {
                Node *nodi = ke->teN[ii].second;
                R(nodi->row) -= be(ii);
            }
        });
    }

    // Apply Dirichlet BCs
    for (auto dBC : pbl->dBCs)
        for (auto nod : dBC->nodes)
            R(nod->row) = 0.;

    if (verbose > 2)
        std::cout << "R (" << R.size() << ")\n";
}

/**
 * @brief Find upwind element which is best aligned with current velocity vector
 */
void Newton::findUpwd()
{
    tbb::parallel_for_each(pbl->fluid->adjMap.begin(), pbl->fluid->adjMap.end(), [&](std::pair<Element *, std::vector<Element *>> &p) {
        // Opposite velocity vector
        Eigen::VectorXd vel_ = -p.first->computeGradient(phi, 0);
        vel_.normalize();
        // Best alignement strategy
        // -> find (cgU-cgE) \dot vel_ closest to one
        double SPbest = -1;
        size_t idx = 0;
        Eigen::VectorXd cgE = (p.first)->getCg().middleRows(0, pbl->nDim);
        for (size_t i = 0; i < p.second.size(); ++i)
        {
            Eigen::VectorXd cgU = (p.second[i])->getCg().middleRows(0, pbl->nDim);
            double SP = (cgU - cgE).normalized().dot(vel_);
            if (fabs(SP - 1) < fabs(SPbest - 1))
            {
                idx = i;
                SPbest = SP;
            }
        }
        // Swap (best upwind element is moved at first emplacement)
        Element *tmp = p.second[0];
        p.second[0] = p.second[idx];
        p.second[idx] = tmp;
    });
}

void Newton::write(std::ostream &out) const
{
    out << "dart::Newton"
        << "\n";
}
